@@ -28,9 +28,7 @@ std::string write_idx_vis::operator()(const multi_idx& idx) const {
   return idx.idxs_.to_string();
 }
 
-std::string write_idx_vis::operator()(const omni_idx& idx) const {
-  return ":";
-}
+std::string write_idx_vis::operator()(const omni_idx& idx) const { return ":"; }
 
 std::string write_idx_vis::operator()(const ub_idx& idx) const {
   std::stringstream ss;
