@@ -1,43 +1,39 @@
-/*
- * -----------------------------------------------------------------
- * $Revision$
- * $Date$
- * -----------------------------------------------------------------
+/* -----------------------------------------------------------------
  * Programmer(s): Alan C. Hindmarsh and Radu Serban @ LLNL
  * -----------------------------------------------------------------
- * LLNS Copyright Start
- * Copyright (c) 2014, Lawrence Livermore National Security
- * This work was performed under the auspices of the U.S. Department 
- * of Energy by Lawrence Livermore National Laboratory in part under 
- * Contract W-7405-Eng-48 and in part under Contract DE-AC52-07NA27344.
- * Produced at the Lawrence Livermore National Laboratory.
+ * SUNDIALS Copyright Start
+ * Copyright (c) 2002-2020, Lawrence Livermore National Security
+ * and Southern Methodist University.
  * All rights reserved.
- * For details, see the LICENSE file.
- * LLNS Copyright End
+ *
+ * See the top-level LICENSE and NOTICE files for details.
+ *
+ * SPDX-License-Identifier: BSD-3-Clause
+ * SUNDIALS Copyright End
  * -----------------------------------------------------------------
  * This is the implementation file for the optional input and output
  * functions for the CVODES solver.
- * -----------------------------------------------------------------
- */
+ * ----------------------------------------------------------------- */
 
 #include <stdio.h>
 #include <stdlib.h>
 
 #include "cvodes_impl.h"
+#include "sundials/sundials_types.h"
+#include "sundials/sundials_math.h"
 
-#include <sundials/sundials_math.h>
-#include <sundials/sundials_types.h>
+#define ZERO   RCONST(0.0)
+#define HALF   RCONST(0.5)
+#define ONE    RCONST(1.0)
+#define TWOPT5 RCONST(2.5)
 
-#define ZERO RCONST(0.0)
-#define ONE  RCONST(1.0)
-
-/* 
+/*
  * =================================================================
  * CVODES optional input functions
  * =================================================================
  */
 
-/* 
+/*
  * CVodeSetErrHandlerFn
  *
  * Specifies the error handler function
@@ -60,7 +56,7 @@ int CVodeSetErrHandlerFn(void *cvode_mem, CVErrHandlerFn ehfun, void *eh_data)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetErrFile
  *
  * Specifies the FILE pointer for output (NULL means no messages)
@@ -82,34 +78,7 @@ int CVodeSetErrFile(void *cvode_mem, FILE *errfp)
   return(CV_SUCCESS);
 }
 
-/* 
- * CVodeSetIterType
- *
- * Specifies the iteration type (CV_FUNCTIONAL or CV_NEWTON)
- */
-
-int CVodeSetIterType(void *cvode_mem, int iter)
-{
-  CVodeMem cv_mem;
-
-  if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetIterType", MSGCV_NO_MEM);
-    return(CV_MEM_NULL);
-  }
-
-  cv_mem = (CVodeMem) cvode_mem;
-
-  if ((iter != CV_FUNCTIONAL) && (iter != CV_NEWTON)) {
-    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetIterType", MSGCV_BAD_ITER);
-    return (CV_ILL_INPUT);
-  }
-
-  cv_mem->cv_iter = iter;
-
-  return(CV_SUCCESS);
-}
-
-/* 
+/*
  * CVodeSetUserData
  *
  * Specifies the user data pointer for f
@@ -131,7 +100,7 @@ int CVodeSetUserData(void *cvode_mem, void *user_data)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxOrd
  *
  * Specifies the maximum method order
@@ -153,7 +122,7 @@ int CVodeSetMaxOrd(void *cvode_mem, int maxord)
     cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetMaxOrd", MSGCV_NEG_MAXORD);
     return(CV_ILL_INPUT);
   }
-  
+
   /* Cannot increase maximum order beyond the value that
      was used when allocating memory */
   qmax_alloc = cv_mem->cv_qmax_alloc;
@@ -170,7 +139,7 @@ int CVodeSetMaxOrd(void *cvode_mem, int maxord)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxNumSteps
  *
  * Specifies the maximum number of integration steps
@@ -196,7 +165,7 @@ int CVodeSetMaxNumSteps(void *cvode_mem, long int mxsteps)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxHnilWarns
  *
  * Specifies the maximum number of warnings for small h
@@ -218,7 +187,7 @@ int CVodeSetMaxHnilWarns(void *cvode_mem, int mxhnil)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  *CVodeSetStabLimDet
  *
  * Turns on/off the stability limit detection algorithm
@@ -245,7 +214,7 @@ int CVodeSetStabLimDet(void *cvode_mem, booleantype sldet)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetInitStep
  *
  * Specifies the initial step size
@@ -267,7 +236,7 @@ int CVodeSetInitStep(void *cvode_mem, realtype hin)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMinStep
  *
  * Specifies the minimum step size
@@ -305,7 +274,7 @@ int CVodeSetMinStep(void *cvode_mem, realtype hmin)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxStep
  *
  * Specifies the maximum step size
@@ -345,7 +314,7 @@ int CVodeSetMaxStep(void *cvode_mem, realtype hmax)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetStopTime
  *
  * Specifies the time beyond which the integration is not to proceed.
@@ -380,7 +349,7 @@ int CVodeSetStopTime(void *cvode_mem, realtype tstop)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxErrTestFails
  *
  * Specifies the maximum number of error test failures during one
@@ -403,10 +372,10 @@ int CVodeSetMaxErrTestFails(void *cvode_mem, int maxnef)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxConvFails
  *
- * Specifies the maximum number of nonlinear convergence failures 
+ * Specifies the maximum number of nonlinear convergence failures
  * during one step try.
  */
 
@@ -426,7 +395,7 @@ int CVodeSetMaxConvFails(void *cvode_mem, int maxncf)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetMaxNonlinIters
  *
  * Specifies the maximum number of nonlinear iterations during
@@ -436,20 +405,46 @@ int CVodeSetMaxConvFails(void *cvode_mem, int maxncf)
 int CVodeSetMaxNonlinIters(void *cvode_mem, int maxcor)
 {
   CVodeMem cv_mem;
+  booleantype sensi_sim;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetMaxNonlinIters", MSGCV_NO_MEM);
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeSetMaxNonlinIters", MSGCV_NO_MEM);
     return (CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
-  cv_mem->cv_maxcor = maxcor;
+  /* Are we computing sensitivities with the simultaneous approach? */
+  sensi_sim = (cv_mem->cv_sensi && (cv_mem->cv_ism==CV_SIMULTANEOUS));
+
+  if (sensi_sim) {
+
+    /* check that the NLS is non-NULL */
+    if (cv_mem->NLSsim == NULL) {
+      cvProcessError(NULL, CV_MEM_FAIL, "CVODES",
+                      "CVodeSetMaxNonlinIters", MSGCV_MEM_FAIL);
+      return(CV_MEM_FAIL);
+    }
+
+    return(SUNNonlinSolSetMaxIters(cv_mem->NLSsim, maxcor));
+
+  } else {
+
+    /* check that the NLS is non-NULL */
+    if (cv_mem->NLS == NULL) {
+      cvProcessError(NULL, CV_MEM_FAIL, "CVODES",
+                      "CVodeSetMaxNonlinIters", MSGCV_MEM_FAIL);
+      return(CV_MEM_FAIL);
+    }
+
+    return(SUNNonlinSolSetMaxIters(cv_mem->NLS, maxcor));
+  }
 
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeSetNonlinConvCoef
  *
  * Specifies the coeficient in the nonlinear solver convergence
@@ -472,7 +467,39 @@ int CVodeSetNonlinConvCoef(void *cvode_mem, realtype nlscoef)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
+ * CVodeSetLSetupFrequency
+ *
+ * Specifies the frequency for calling the linear solver setup function to
+ * recompute the Jacobian matrix and/or preconditioner
+ */
+
+int CVodeSetLSetupFrequency(void *cvode_mem, long int msbp)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem == NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetLSetupFrequency",
+                   MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  /* check for a valid input */
+  if (msbp < 0) {
+    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetLSetupFrequency",
+                   "A negative setup frequency was provided");
+    return(CV_ILL_INPUT);
+  }
+
+  /* use default or user provided value */
+  cv_mem->cv_msbp = (msbp == 0) ? MSBP : msbp;
+
+  return(CV_SUCCESS);
+}
+
+/*
  * CVodeSetRootDirection
  *
  * Specifies the direction of zero-crossings to be monitored.
@@ -494,7 +521,7 @@ int CVodeSetRootDirection(void *cvode_mem, int *rootdir)
   nrt = cv_mem->cv_nrtfn;
   if (nrt==0) {
     cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetRootDirection", MSGCV_NO_ROOT);
-    return(CV_ILL_INPUT);    
+    return(CV_ILL_INPUT);
   }
 
   for(i=0; i<nrt; i++) cv_mem->cv_rootdir[i] = rootdir[i];
@@ -522,11 +549,74 @@ int CVodeSetNoInactiveRootWarn(void *cvode_mem)
   cv_mem = (CVodeMem) cvode_mem;
 
   cv_mem->cv_mxgnull = 0;
-  
+
   return(CV_SUCCESS);
 }
 
-/* 
+/*
+ * CVodeSetConstraints
+ *
+ * Setup for constraint handling feature
+ */
+
+int CVodeSetConstraints(void *cvode_mem, N_Vector constraints)
+{
+  CVodeMem cv_mem;
+  realtype temptest;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetConstraints", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  /* If there are no constraints, destroy data structures */
+  if (constraints == NULL) {
+    if (cv_mem->cv_constraintsMallocDone) {
+      N_VDestroy(cv_mem->cv_constraints);
+      cv_mem->cv_lrw -= cv_mem->cv_lrw1;
+      cv_mem->cv_liw -= cv_mem->cv_liw1;
+    }
+    cv_mem->cv_constraintsMallocDone = SUNFALSE;
+    cv_mem->cv_constraintsSet = SUNFALSE;
+    return(CV_SUCCESS);
+  }
+
+  /* Test if required vector ops. are defined */
+
+  if (constraints->ops->nvdiv         == NULL ||
+      constraints->ops->nvmaxnorm     == NULL ||
+      constraints->ops->nvcompare     == NULL ||
+      constraints->ops->nvconstrmask  == NULL ||
+      constraints->ops->nvminquotient == NULL) {
+    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetConstraints", MSGCV_BAD_NVECTOR);
+    return(CV_ILL_INPUT);
+  }
+
+  /* Check the constraints vector */
+  temptest = N_VMaxNorm(constraints);
+  if ((temptest > TWOPT5) || (temptest < HALF)) {
+    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetConstraints", MSGCV_BAD_CONSTR);
+    return(CV_ILL_INPUT);
+  }
+
+  if ( !(cv_mem->cv_constraintsMallocDone) ) {
+    cv_mem->cv_constraints = N_VClone(constraints);
+    cv_mem->cv_lrw += cv_mem->cv_lrw1;
+    cv_mem->cv_liw += cv_mem->cv_liw1;
+    cv_mem->cv_constraintsMallocDone = SUNTRUE;
+  }
+
+  /* Load the constraints vector */
+  N_VScale(ONE, constraints, cv_mem->cv_constraints);
+
+  cv_mem->cv_constraintsSet = SUNTRUE;
+
+  return(CV_SUCCESS);
+}
+
+/*
  * =================================================================
  * Quadrature optional input functions
  * =================================================================
@@ -537,7 +627,7 @@ int CVodeSetQuadErrCon(void *cvode_mem, booleantype errconQ)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetQuadErrCon", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetQuadErrCon", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
   cv_mem = (CVodeMem) cvode_mem;
@@ -547,7 +637,7 @@ int CVodeSetQuadErrCon(void *cvode_mem, booleantype errconQ)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * =================================================================
  * FSA optional input functions
  * =================================================================
@@ -558,19 +648,19 @@ int CVodeSetSensDQMethod(void *cvode_mem, int DQtype, realtype DQrhomax)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensDQMethod", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensDQMethod", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if ( (DQtype != CV_CENTERED) && (DQtype != CV_FORWARD) ) {
-    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetSensDQMethod", MSGCV_BAD_DQTYPE);    
+    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetSensDQMethod", MSGCV_BAD_DQTYPE);
     return(CV_ILL_INPUT);
   }
 
   if (DQrhomax < ZERO ) {
-    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetSensDQMethod", MSGCV_BAD_DQRHO);    
+    cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES", "CVodeSetSensDQMethod", MSGCV_BAD_DQRHO);
     return(CV_ILL_INPUT);
   }
 
@@ -587,7 +677,7 @@ int CVodeSetSensErrCon(void *cvode_mem, booleantype errconS)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensErrCon", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensErrCon", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
   cv_mem = (CVodeMem) cvode_mem;
@@ -602,15 +692,41 @@ int CVodeSetSensErrCon(void *cvode_mem, booleantype errconS)
 int CVodeSetSensMaxNonlinIters(void *cvode_mem, int maxcorS)
 {
   CVodeMem cv_mem;
+  booleantype sensi_stg;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensMaxNonlinIters", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeSetSensMaxNonlinIters", MSGCV_NO_MEM);
     return (CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
-  cv_mem->cv_maxcorS = maxcorS;
+  /* Are we computing sensitivities with a staggered approach? */
+  sensi_stg = (cv_mem->cv_sensi && (cv_mem->cv_ism==CV_STAGGERED));
+
+  if (sensi_stg) {
+
+    /* check that the NLS is non-NULL */
+    if (cv_mem->NLSstg == NULL) {
+      cvProcessError(NULL, CV_MEM_FAIL, "CVODES",
+                      "CVodeSetSensMaxNonlinIters", MSGCV_MEM_FAIL);
+      return(CV_MEM_FAIL);
+    }
+
+    return(SUNNonlinSolSetMaxIters(cv_mem->NLSstg, maxcorS));
+
+  } else {
+
+    /* check that the NLS is non-NULL */
+    if (cv_mem->NLSstg1 == NULL) {
+      cvProcessError(NULL, CV_MEM_FAIL, "CVODES",
+                      "CVodeSetMaxNonlinIters", MSGCV_MEM_FAIL);
+      return(CV_MEM_FAIL);
+    }
+
+    return(SUNNonlinSolSetMaxIters(cv_mem->NLSstg1, maxcorS));
+  }
 
   return(CV_SUCCESS);
 }
@@ -623,7 +739,7 @@ int CVodeSetSensParams(void *cvode_mem, realtype *p, realtype *pbar, int *plist)
   int is, Ns;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensParams", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetSensParams", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -682,7 +798,7 @@ int CVodeSetQuadSensErrCon(void *cvode_mem, booleantype errconQS)
   if (cvode_mem==NULL) {
     cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeSetQuadSensErrCon", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
-  }  
+  }
   cv_mem = (CVodeMem) cvode_mem;
 
   /* Was sensitivity initialized? */
@@ -690,12 +806,12 @@ int CVodeSetQuadSensErrCon(void *cvode_mem, booleantype errconQS)
   if (cv_mem->cv_SensMallocDone == SUNFALSE) {
     cvProcessError(cv_mem, CV_NO_SENS, "CVODES", "CVodeSetQuadSensTolerances", MSGCV_NO_SENSI);
     return(CV_NO_SENS);
-  } 
+  }
 
   /* Ckeck if quadrature sensitivity was initialized? */
 
   if (cv_mem->cv_QuadSensMallocDone == SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeSetQuadSensErrCon", MSGCV_NO_QUADSENSI); 
+    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeSetQuadSensErrCon", MSGCV_NO_QUADSENSI);
     return(CV_NO_QUAD);
   }
 
@@ -704,7 +820,7 @@ int CVodeSetQuadSensErrCon(void *cvode_mem, booleantype errconQS)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * =================================================================
  * CVODES optional output functions
  * =================================================================
@@ -732,7 +848,7 @@ int CVodeGetNumSteps(void *cvode_mem, long int *nsteps)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetNumRhsEvals
  *
  * Returns the current number of calls to f
@@ -754,7 +870,7 @@ int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevals)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetNumLinSolvSetups
  *
  * Returns the current number of calls to the linear solver setup routine
@@ -798,7 +914,7 @@ int CVodeGetNumErrTestFails(void *cvode_mem, long int *netfails)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetLastOrder
  *
  * Returns the order on the last succesful step
@@ -820,7 +936,7 @@ int CVodeGetLastOrder(void *cvode_mem, int *qlast)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetCurrentOrder
  *
  * Returns the order to be attempted on the next step
@@ -842,7 +958,29 @@ int CVodeGetCurrentOrder(void *cvode_mem, int *qcur)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
+ * CVodeGetCurrentGamma
+ *
+ * Returns the value of gamma for the current step.
+ */
+
+int CVodeGetCurrentGamma(void *cvode_mem, realtype *gamma)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetCurrentGamma", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  *gamma = cv_mem->cv_gamma;
+
+  return(CV_SUCCESS);
+}
+
+/*
  * CVodeGetNumStabLimOrderReds
  *
  * Returns the number of order reductions triggered by the stability
@@ -868,7 +1006,7 @@ int CVodeGetNumStabLimOrderReds(void *cvode_mem, long int *nslred)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetActualInitStep
  *
  * Returns the step size used on the first step
@@ -912,7 +1050,7 @@ int CVodeGetLastStep(void *cvode_mem, realtype *hlast)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetCurrentStep
  *
  * Returns the step size to be attempted on the next step
@@ -928,13 +1066,80 @@ int CVodeGetCurrentStep(void *cvode_mem, realtype *hcur)
   }
 
   cv_mem = (CVodeMem) cvode_mem;
-  
+
   *hcur = cv_mem->cv_next_h;
 
   return(CV_SUCCESS);
 }
 
-/* 
+/*
+ * CVodeGetCurrentState
+ *
+ * Returns the current state vector
+ */
+
+int CVodeGetCurrentState(void *cvode_mem, N_Vector *y)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetCurrentState", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  *y = cv_mem->cv_y;
+
+  return(CV_SUCCESS);
+}
+
+/*
+ * CVodeGetCurrentStateSens
+ *
+ * Returns the current sensitivity state vector array
+ */
+
+int CVodeGetCurrentStateSens(void *cvode_mem, N_Vector **yS)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetCurrentStateSens", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  *yS = cv_mem->cv_yS;
+
+  return(CV_SUCCESS);
+}
+
+/*
+ * CVodeGetCurrentSensSolveIndex
+ *
+ * Returns the current index of the sensitivity solve when using
+ * the staggered1 nonlinear solver.
+ */
+
+int CVodeGetCurrentSensSolveIndex(void *cvode_mem, int *index)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetCurrentSensSolveIndex", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  *index = cv_mem->sens_solve_idx;
+
+  return(CV_SUCCESS);
+}
+
+/*
  * CVodeGetCurrentTime
  *
  * Returns the current value of the independent variable
@@ -956,7 +1161,7 @@ int CVodeGetCurrentTime(void *cvode_mem, realtype *tcur)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetTolScaleFactor
  *
  * Returns a suggested factor for scaling tolerances
@@ -978,7 +1183,7 @@ int CVodeGetTolScaleFactor(void *cvode_mem, realtype *tolsfact)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetErrWeights
  *
  * This routine returns the current weight vector.
@@ -1022,7 +1227,7 @@ int CVodeGetEstLocalErrors(void *cvode_mem, N_Vector ele)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetWorkSpace
  *
  * Returns integrator work space requirements
@@ -1045,15 +1250,15 @@ int CVodeGetWorkSpace(void *cvode_mem, long int *lenrw, long int *leniw)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetIntegratorStats
  *
  * Returns integrator statistics
  */
 
-int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps, long int *nfevals, 
-                            long int *nlinsetups, long int *netfails, int *qlast, 
-                            int *qcur, realtype *hinused, realtype *hlast, 
+int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps, long int *nfevals,
+                            long int *nlinsetups, long int *netfails, int *qlast,
+                            int *qcur, realtype *hinused, realtype *hlast,
                             realtype *hcur, realtype *tcur)
 {
   CVodeMem cv_mem;
@@ -1079,7 +1284,7 @@ int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps, long int *nfevals
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetNumGEvals
  *
  * Returns the current number of calls to g (for rootfinding)
@@ -1101,7 +1306,7 @@ int CVodeGetNumGEvals(void *cvode_mem, long int *ngevals)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetRootInfo
  *
  * Returns pointer to array rootsfound showing roots found
@@ -1127,7 +1332,7 @@ int CVodeGetRootInfo(void *cvode_mem, int *rootsfound)
 }
 
 
-/* 
+/*
  * CVodeGetNumNonlinSolvIters
  *
  * Returns the current number of iterations in the nonlinear solver
@@ -1138,7 +1343,8 @@ int CVodeGetNumNonlinSolvIters(void *cvode_mem, long int *nniters)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetNumNonlinSolvIters", MSGCV_NO_MEM);
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeGetNumNonlinSolvIters", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1149,7 +1355,7 @@ int CVodeGetNumNonlinSolvIters(void *cvode_mem, long int *nniters)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetNumNonlinSolvConvFails
  *
  * Returns the current number of convergence failures in the
@@ -1172,32 +1378,33 @@ int CVodeGetNumNonlinSolvConvFails(void *cvode_mem, long int *nncfails)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * CVodeGetNonlinSolvStats
  *
  * Returns nonlinear solver statistics
  */
 
-int CVodeGetNonlinSolvStats(void *cvode_mem, long int *nniters, 
+int CVodeGetNonlinSolvStats(void *cvode_mem, long int *nniters,
                             long int *nncfails)
 {
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetNonlinSolvStats", MSGCV_NO_MEM);
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeGetNonlinSolvStats", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
-  *nniters = cv_mem->cv_nni;
+  *nniters  = cv_mem->cv_nni;
   *nncfails = cv_mem->cv_ncfn;
 
   return(CV_SUCCESS);
 }
 
 
-/* 
+/*
  * =================================================================
  * Quadrature optional output functions
  * =================================================================
@@ -1210,14 +1417,14 @@ int CVodeGetQuadNumRhsEvals(void *cvode_mem, long int *nfQevals)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadNumRhsEvals", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadNumRhsEvals", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadNumRhsEvals", MSGCV_NO_QUAD); 
+    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadNumRhsEvals", MSGCV_NO_QUAD);
     return(CV_NO_QUAD);
   }
 
@@ -1233,14 +1440,14 @@ int CVodeGetQuadNumErrTestFails(void *cvode_mem, long int *nQetfails)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadNumErrTestFails", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadNumErrTestFails", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadNumErrTestFails", MSGCV_NO_QUAD); 
+    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadNumErrTestFails", MSGCV_NO_QUAD);
     return(CV_NO_QUAD);
   }
 
@@ -1256,14 +1463,14 @@ int CVodeGetQuadErrWeights(void *cvode_mem, N_Vector eQweight)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadErrWeights", MSGCV_NO_MEM); 
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadErrWeights", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadErrWeights", MSGCV_NO_QUAD); 
+    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadErrWeights", MSGCV_NO_QUAD);
     return(CV_NO_QUAD);
   }
 
@@ -1279,14 +1486,14 @@ int CVodeGetQuadStats(void *cvode_mem, long int *nfQevals, long int *nQetfails)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadStats", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadStats", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadStats", MSGCV_NO_QUAD); 
+    cvProcessError(cv_mem, CV_NO_QUAD, "CVODES", "CVodeGetQuadStats", MSGCV_NO_QUAD);
     return(CV_NO_QUAD);
   }
 
@@ -1296,7 +1503,7 @@ int CVodeGetQuadStats(void *cvode_mem, long int *nfQevals, long int *nQetfails)
   return(CV_SUCCESS);
 }
 
-/* 
+/*
  * =================================================================
  * Quadrature FSA optional output functions
  * =================================================================
@@ -1309,14 +1516,14 @@ int CVodeGetQuadSensNumRhsEvals(void *cvode_mem, long int *nfQSevals)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensNumRhsEvals", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensNumRhsEvals", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr_sensi == SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensNumRhsEvals", MSGCV_NO_QUADSENSI); 
+    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensNumRhsEvals", MSGCV_NO_QUADSENSI);
     return(CV_NO_QUADSENS);
   }
 
@@ -1332,14 +1539,14 @@ int CVodeGetQuadSensNumErrTestFails(void *cvode_mem, long int *nQSetfails)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensNumErrTestFails", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensNumErrTestFails", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr_sensi == SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensNumErrTestFails", MSGCV_NO_QUADSENSI); 
+    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensNumErrTestFails", MSGCV_NO_QUADSENSI);
     return(CV_NO_QUADSENS);
   }
 
@@ -1356,14 +1563,14 @@ int CVodeGetQuadSensErrWeights(void *cvode_mem, N_Vector *eQSweight)
   int is, Ns;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensErrWeights", MSGCV_NO_MEM); 
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensErrWeights", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr_sensi == SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensErrWeights", MSGCV_NO_QUADSENSI); 
+    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensErrWeights", MSGCV_NO_QUADSENSI);
     return(CV_NO_QUADSENS);
   }
   Ns = cv_mem->cv_Ns;
@@ -1382,14 +1589,14 @@ int CVodeGetQuadSensStats(void *cvode_mem, long int *nfQSevals, long int *nQSetf
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensStats", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetQuadSensStats", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_quadr_sensi == SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensStats", MSGCV_NO_QUADSENSI); 
+    cvProcessError(cv_mem, CV_NO_QUADSENS, "CVODES", "CVodeGetQuadSensStats", MSGCV_NO_QUADSENSI);
     return(CV_NO_QUADSENS);
   }
 
@@ -1400,7 +1607,7 @@ int CVodeGetQuadSensStats(void *cvode_mem, long int *nfQSevals, long int *nQSetf
 }
 
 
-/* 
+/*
  * =================================================================
  * FSA optional output functions
  * =================================================================
@@ -1413,7 +1620,7 @@ int CVodeGetSensNumRhsEvals(void *cvode_mem, long int *nfSevals)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumRhsEvals", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumRhsEvals", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1436,7 +1643,7 @@ int CVodeGetNumRhsEvalsSens(void *cvode_mem, long int *nfevalsS)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetNumRhsEvalsSens", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetNumRhsEvalsSens", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1459,7 +1666,7 @@ int CVodeGetSensNumErrTestFails(void *cvode_mem, long int *nSetfails)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumErrTestFails", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumErrTestFails", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1482,7 +1689,7 @@ int CVodeGetSensNumLinSolvSetups(void *cvode_mem, long int *nlinsetupsS)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumLinSolvSetups", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumLinSolvSetups", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1506,7 +1713,7 @@ int CVodeGetSensErrWeights(void *cvode_mem, N_Vector *eSweight)
   int is, Ns;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensErrWeights", MSGCV_NO_MEM);    
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensErrWeights", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
@@ -1527,7 +1734,7 @@ int CVodeGetSensErrWeights(void *cvode_mem, N_Vector *eSweight)
 
 /*-----------------------------------------------------------------*/
 
-int CVodeGetSensStats(void *cvode_mem, long int *nfSevals, long int *nfevalsS, 
+int CVodeGetSensStats(void *cvode_mem, long int *nfSevals, long int *nfevalsS,
                       long int *nSetfails, long int *nlinsetupsS)
 {
   CVodeMem cv_mem;
@@ -1559,14 +1766,16 @@ int CVodeGetSensNumNonlinSolvIters(void *cvode_mem, long int *nSniters)
   CVodeMem cv_mem;
 
   if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNumNonlinSolvIters", MSGCV_NO_MEM);
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeGetSensNumNonlinSolvIters", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
   if (cv_mem->cv_sensi==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_SENS, "CVODES", "CVodeGetSensNumNonlinSolvIters", MSGCV_NO_SENSI);
+    cvProcessError(cv_mem, CV_NO_SENS, "CVODES",
+                   "CVodeGetSensNumNonlinSolvIters", MSGCV_NO_SENSI);
     return(CV_NO_SENS);
   }
 
@@ -1600,6 +1809,33 @@ int CVodeGetSensNumNonlinSolvConvFails(void *cvode_mem, long int *nSncfails)
 
 /*-----------------------------------------------------------------*/
 
+int CVodeGetSensNonlinSolvStats(void *cvode_mem, long int *nSniters,
+                                long int *nSncfails)
+{
+  CVodeMem cv_mem;
+
+  if (cvode_mem==NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeGetSensNonlinSolvStats", MSGCV_NO_MEM);
+    return(CV_MEM_NULL);
+  }
+
+  cv_mem = (CVodeMem) cvode_mem;
+
+  if (cv_mem->cv_sensi==SUNFALSE) {
+    cvProcessError(cv_mem, CV_NO_SENS, "CVODES",
+                   "CVodeGetSensNonlinSolvStats", MSGCV_NO_SENSI);
+    return(CV_NO_SENS);
+  }
+
+  *nSniters  = cv_mem->cv_nniS;
+  *nSncfails = cv_mem->cv_ncfnS;
+
+  return(CV_SUCCESS);
+}
+
+/*-----------------------------------------------------------------*/
+
 int CVodeGetStgrSensNumNonlinSolvIters(void *cvode_mem, long int *nSTGR1niters)
 {
   CVodeMem cv_mem;
@@ -1619,7 +1855,7 @@ int CVodeGetStgrSensNumNonlinSolvIters(void *cvode_mem, long int *nSTGR1niters)
     return(CV_NO_SENS);
   }
 
-  if(cv_mem->cv_ism==CV_STAGGERED1) 
+  if(cv_mem->cv_ism==CV_STAGGERED1)
     for(is=0; is<Ns; is++) nSTGR1niters[is] = cv_mem->cv_nniS1[is];
 
   return(CV_SUCCESS);
@@ -1646,7 +1882,7 @@ int CVodeGetStgrSensNumNonlinSolvConvFails(void *cvode_mem, long int *nSTGR1ncfa
     return(CV_NO_SENS);
   }
 
-  if(cv_mem->cv_ism==CV_STAGGERED1) 
+  if(cv_mem->cv_ism==CV_STAGGERED1)
     for(is=0; is<Ns; is++) nSTGR1ncfails[is] = cv_mem->cv_ncfnS1[is];
 
   return(CV_SUCCESS);
@@ -1654,25 +1890,33 @@ int CVodeGetStgrSensNumNonlinSolvConvFails(void *cvode_mem, long int *nSTGR1ncfa
 
 /*-----------------------------------------------------------------*/
 
-int CVodeGetSensNonlinSolvStats(void *cvode_mem, long int *nSniters, 
-                                long int *nSncfails)
+int CVodeGetStgrSensNonlinSolvStats(void *cvode_mem,
+                                    long int *nSTGR1niters,
+                                    long int *nSTGR1ncfails)
 {
   CVodeMem cv_mem;
+  int is, Ns;
 
-  if (cvode_mem==NULL) {
-    cvProcessError(NULL, CV_MEM_NULL, "CVODES", "CVodeGetSensNonlinSolvstats", MSGCV_NO_MEM);
+  if (cvode_mem == NULL) {
+    cvProcessError(NULL, CV_MEM_NULL, "CVODES",
+                   "CVodeGetStgrSensNonlinSolvStats", MSGCV_NO_MEM);
     return(CV_MEM_NULL);
   }
 
   cv_mem = (CVodeMem) cvode_mem;
 
-  if (cv_mem->cv_sensi==SUNFALSE) {
-    cvProcessError(cv_mem, CV_NO_SENS, "CVODES", "CVodeGetSensNonlinSolvStats", MSGCV_NO_SENSI);
+  Ns = cv_mem->cv_Ns;
+
+  if (cv_mem->cv_sensi == SUNFALSE) {
+    cvProcessError(cv_mem, CV_NO_SENS, "CVODES",
+                   "CVodeGetStgrSensNonlinSolvStats", MSGCV_NO_SENSI);
     return(CV_NO_SENS);
   }
 
-  *nSniters = cv_mem->cv_nniS;
-  *nSncfails = cv_mem->cv_ncfnS;
+  if(cv_mem->cv_ism == CV_STAGGERED1) {
+    for(is=0; is<Ns; is++) nSTGR1niters[is]  = cv_mem->cv_nniS1[is];
+    for(is=0; is<Ns; is++) nSTGR1ncfails[is] = cv_mem->cv_ncfnS1[is];
+  }
 
   return(CV_SUCCESS);
 }
@@ -1787,7 +2031,7 @@ char *CVodeGetReturnFlagName(long int flag)
     break;
   case CV_TOO_CLOSE:
     sprintf(name,"CV_TOO_CLOSE");
-    break; 
+    break;
   case CV_NO_ADJ:
     sprintf(name,"CV_NO_ADJ");
     break;
@@ -1808,11 +2052,13 @@ char *CVodeGetReturnFlagName(long int flag)
     break;
   case CV_GETY_BADT:
     sprintf(name,"CV_GETY_BADT");
-    break;  
+    break;
+  case CV_NLS_FAIL:
+    sprintf(name,"CV_NLS_FAIL");
+    break;
   default:
     sprintf(name,"NONE");
   }
 
   return(name);
 }
-
